#include "OpenSPC.h"
#include "config.h"
#include "tty.h"

static int tty_fd = 0;
static struct termios saved_ti;

void tty_init(void)
{
	int fd;
	struct termios ti;

	if (tty_fd) return;

	if ((fd = open("/dev/tty", O_RDONLY | O_NONBLOCK)) < 0) return;

	if (tcgetpgrp(fd) != getpid()) {
		close(fd); return;
	}

	tcgetattr(fd, &ti);
	saved_ti = ti;
	ti.c_lflag &= ~(ICANON | ECHO);
	ti.c_cc[VMIN] = 1;
	ti.c_cc[VTIME] = 0;
	tcsetattr(fd, TCSANOW, &ti);

	tty_fd = fd;

	atexit(tty_done);
}

int tty_getchar(void)
{
	int c;

	if (tty_fd) {
		c = 0;
		if (read(tty_fd, &c, 1) > 0) return c;
	}

	return -1;
}

void tty_done(void)
{
	int fd;

	if (!tty_fd) return;

	fd = tty_fd; tty_fd = 0;
	tcsetattr(fd, TCSANOW, &saved_ti);

	close(fd);
}
